@@ -1,4 +1,4 @@
-/* Copyright 1995-1996,1997-2014,2018
+/* Copyright 1995-2014,2016,2018-2020
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -32,6 +32,7 @@
 #include "keywords.h"
 #include "list.h"
 #include "macros.h"
+#include "memoize.h"    /* for the SCM_API declarations of 'scm_sym_' */
 #include "modules.h"
 #include "pairs.h"
 #include "ports.h"
@@ -74,12 +75,12 @@ static const char** exp_field_names[SCM_NUM_EXPANDED_TYPES];
   SCM_MAKE_EXPANDED_MODULE_REF(src, mod, name, public)
 #define MODULE_SET(src, mod, name, public, exp) \
   SCM_MAKE_EXPANDED_MODULE_SET(src, mod, name, public, exp)
-#define TOPLEVEL_REF(src, name) \
-  SCM_MAKE_EXPANDED_TOPLEVEL_REF(src, name)
-#define TOPLEVEL_SET(src, name, exp) \
-  SCM_MAKE_EXPANDED_TOPLEVEL_SET(src, name, exp)
-#define TOPLEVEL_DEFINE(src, name, exp) \
-  SCM_MAKE_EXPANDED_TOPLEVEL_DEFINE(src, name, exp)
+#define TOPLEVEL_REF(src, mod, name) \
+  SCM_MAKE_EXPANDED_TOPLEVEL_REF(src, mod, name)
+#define TOPLEVEL_SET(src, mod, name, exp) \
+  SCM_MAKE_EXPANDED_TOPLEVEL_SET(src, mod, name, exp)
+#define TOPLEVEL_DEFINE(src, mod, name, exp) \
+  SCM_MAKE_EXPANDED_TOPLEVEL_DEFINE(src, mod, name, exp)
 #define CONDITIONAL(src, test, consequent, alternate) \
   SCM_MAKE_EXPANDED_CONDITIONAL(src, test, consequent, alternate)
 #define PRIMCALL(src, name, exps) \
@@ -224,7 +225,7 @@ SCM_KEYWORD (kw_rest, "rest");
 static void 
 syntax_error (const char* const msg, const SCM form, const SCM expr)
 {
-  SCM msg_string = scm_from_locale_string (msg);
+  SCM msg_string = scm_from_utf8_string (msg);
   SCM filename = SCM_BOOL_F;
   SCM linenr = SCM_BOOL_F;
   const char *format;
@@ -377,7 +378,7 @@ expand (SCM exp, SCM env)
       if (scm_is_true (gensym))
         return LEXICAL_REF (SCM_BOOL_F, exp, gensym);
       else
-        return TOPLEVEL_REF (SCM_BOOL_F, exp);
+        return TOPLEVEL_REF (SCM_BOOL_F, SCM_BOOL_F, exp);
     }
   else
     return CONST_ (SCM_BOOL_F, exp);
@@ -487,7 +488,7 @@ expand_cond_clauses (SCM clause, SCM rest, int elp, int alp, SCM env)
       && scm_is_eq (CADR (clause), scm_sym_arrow)
       && alp)
     {
-      SCM tmp = scm_gensym (scm_from_locale_string ("cond "));
+      SCM tmp = scm_gensym (scm_from_utf8_string ("cond "));
       SCM new_env = scm_acons (tmp, tmp, env);
       ASSERT_SYNTAX (length > 2, s_missing_recipient, clause);
       ASSERT_SYNTAX (length == 3, s_extra_expression, clause);
@@ -552,13 +553,14 @@ expand_define (SCM expr, SCM env)
       ASSERT_SYNTAX_2 (scm_is_symbol (CAR (variable)), s_bad_variable, variable, expr);
       return TOPLEVEL_DEFINE
         (scm_source_properties (expr),
+         SCM_BOOL_F,
          CAR (variable),
          expand_lambda (scm_cons (scm_sym_lambda, scm_cons (CDR (variable), body)),
                         env));
     }
   ASSERT_SYNTAX_2 (scm_is_symbol (variable), s_bad_variable, variable, expr);
   ASSERT_SYNTAX (scm_ilength (body) == 1, s_expression, expr);
-  return TOPLEVEL_DEFINE (scm_source_properties (expr), variable,
+  return TOPLEVEL_DEFINE (scm_source_properties (expr), SCM_BOOL_F, variable,
                           expand (CAR (body), env));
 }
 
@@ -1143,6 +1145,7 @@ expand_set_x (SCM expr, SCM env)
                           expand (CADDR (expr), env));
     case SCM_EXPANDED_TOPLEVEL_REF:
       return TOPLEVEL_SET (scm_source_properties (expr),
+                           SCM_EXPANDED_REF (vmem, TOPLEVEL_REF, MOD),
                            SCM_EXPANDED_REF (vmem, TOPLEVEL_REF, NAME),
                            expand (CADDR (expr), env));
     case SCM_EXPANDED_MODULE_REF:
@@ -1371,12 +1374,14 @@ convert_assignment (SCM exp, SCM assigned)
     case SCM_EXPANDED_TOPLEVEL_SET:
       return TOPLEVEL_SET
 	 (REF (exp, TOPLEVEL_SET, SRC),
+          REF (exp, TOPLEVEL_SET, MOD),
           REF (exp, TOPLEVEL_SET, NAME),
           convert_assignment (REF (exp, TOPLEVEL_SET, EXP), assigned));
 
     case SCM_EXPANDED_TOPLEVEL_DEFINE:
       return TOPLEVEL_DEFINE
         (REF (exp, TOPLEVEL_DEFINE, SRC),
+         REF (exp, TOPLEVEL_DEFINE, MOD),
          REF (exp, TOPLEVEL_DEFINE, NAME),
          convert_assignment (REF (exp, TOPLEVEL_DEFINE, EXP),
                              assigned));
@@ -1596,7 +1601,7 @@ make_exp_vtable (size_t n)
   
   layout = scm_string_to_symbol
     (scm_string_append (scm_make_list (scm_from_size_t (exp_nfields[n]),
-                                       scm_from_locale_string ("pw"))));
+                                       scm_from_utf8_string ("pw"))));
   printer = SCM_BOOL_F;
   name = scm_from_utf8_symbol (exp_names[n]);
   code = scm_from_size_t (n);
@@ -1638,7 +1643,7 @@ scm_init_expand ()
   DEFINE_NAMES (LETREC);
 
   scm_exp_vtable_vtable =
-    scm_make_vtable (scm_from_locale_string (SCM_VTABLE_BASE_LAYOUT "pwuwpw"),
+    scm_make_vtable (scm_from_utf8_string (SCM_VTABLE_BASE_LAYOUT "pwuwpw"),
                      SCM_BOOL_F);
 
   for (n = 0; n < SCM_NUM_EXPANDED_TYPES; n++)
