@@ -1,4 +1,4 @@
-/* Copyright 1997-2001,2004,2006-2007,2010-2012,2018
+/* Copyright 1997-2001,2004,2006-2007,2010-2012,2018-2019
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -176,7 +176,7 @@ SCM_DEFINE (scm_make_regexp, "make-regexp", 1, 0, 1,
       SCM errmsg = scm_regexp_error_msg (status, rx);
       scm_gc_free (rx, sizeof(regex_t), "regex");
       scm_error_scm (scm_regexp_error_key,
-		     scm_from_locale_string (FUNC_NAME),
+		     scm_from_utf8_string (FUNC_NAME),
 		     errmsg,
 		     SCM_BOOL_F,
 		     scm_list_1 (pat));
@@ -307,7 +307,7 @@ SCM_DEFINE (scm_regexp_exec, "regexp-exec", 2, 2, 0,
 
   if (status != 0 && status != REG_NOMATCH)
     scm_error_scm (scm_regexp_error_key,
-		   scm_from_locale_string (FUNC_NAME),
+		   scm_from_utf8_string (FUNC_NAME),
 		   scm_regexp_error_msg (status, SCM_RGX (rx)),
 		   SCM_BOOL_F, SCM_BOOL_F);
   return mvec;
