@@ -176,16 +176,11 @@
 #define ALLOC_FRAME(n)                                              \
   do {                                                              \
     sp = VP->fp - (n);                                              \
-    if (sp < VP->sp_min_since_gc)                                   \
+    if (SCM_UNLIKELY (sp < VP->stack_limit))                        \
       {                                                             \
-        if (SCM_UNLIKELY (sp < VP->stack_limit))                    \
-          {                                                         \
-            SYNC_IP ();                                             \
-            CALL_INTRINSIC (expand_stack, (thread, sp));            \
-            CACHE_SP ();                                            \
-          }                                                         \
-        else                                                        \
-          VP->sp_min_since_gc = VP->sp = sp;                        \
+        SYNC_IP ();                                                 \
+        CALL_INTRINSIC (expand_stack, (thread, sp));                \
+        CACHE_SP ();                                                \
       }                                                             \
     else                                                            \
       VP->sp = sp;                                                  \
@@ -195,12 +190,7 @@
    stack expansion is needed.  Note that in some cases this may lower
    SP, e.g. after a return but where there are more locals below, but we
    know it was preceded by an alloc-frame in that case, so no stack need
-   be allocated.
-
-   As an optimization, we don't update sp_min_since_gc in this case; the
-   principal place stacks are expanded is in ALLOC_FRAME. it doesn't
-   need to strictly be the min since GC, as it's just an optimization to
-   prevent passing too-large of a range to madvise.  */
+   be allocated.  */
 #define RESET_FRAME(n)                                              \
   do {                                                              \
     VP->sp = sp = VP->fp - (n);                                     \
@@ -3329,9 +3319,63 @@ VM_NAME (scm_thread *thread)
       NEXT (1);
     }
 
-  VM_DEFINE_OP (160, unused_160, NULL, NOP)
-  VM_DEFINE_OP (161, unused_161, NULL, NOP)
-  VM_DEFINE_OP (162, unused_162, NULL, NOP)
+  /* call-scm-scm a:12 b:12 IDX:32
+   *
+   * Call the void-returning instrinsic with index IDX, passing SCM
+   * locals A and B as arguments.
+   */
+  VM_DEFINE_OP (160, call_scm_scm, "call-scm-scm", OP2 (X8_S12_S12, C32))
+    {
+      uint16_t a, b;
+      scm_t_scm_scm_intrinsic intrinsic;
+
+      UNPACK_12_12 (op, a, b);
+      intrinsic = intrinsics[ip[1]];
+
+      SYNC_IP ();
+      intrinsic (SP_REF (a), SP_REF (b));
+
+      NEXT (2);
+    }
+
+  /* call-scm-scm-scm a:8 b:8 c:8 IDX:32
+   *
+   * Call the void-returning instrinsic with index IDX, passing SCM
+   * locals A, B, and C as arguments.
+   */
+  VM_DEFINE_OP (161, call_scm_scm_scm, "call-scm-scm-scm", OP2 (X8_S8_S8_S8, C32))
+    {
+      uint8_t a, b, c;
+      scm_t_scm_scm_scm_intrinsic intrinsic;
+
+      UNPACK_8_8_8 (op, a, b, c);
+      intrinsic = intrinsics[ip[1]];
+
+      SYNC_IP ();
+      intrinsic (SP_REF (a), SP_REF (b), SP_REF (c));
+
+      NEXT (2);
+    }
+
+  /* call-scm-uimm-scm a:8 b:8 c:8 IDX:32
+   *
+   * Call the void-returning instrinsic with index IDX, passing SCM
+   * local A, uint8 B, and SCM local C as arguments.
+   */
+  VM_DEFINE_OP (162, call_scm_uimm_scm, "call-scm-uimm-scm", OP2 (X8_S8_C8_S8, C32))
+    {
+      uint8_t a, b, c;
+      scm_t_scm_uimm_scm_intrinsic intrinsic;
+
+      UNPACK_8_8_8 (op, a, b, c);
+      intrinsic = intrinsics[ip[1]];
+
+      SYNC_IP ();
+      intrinsic (SP_REF (a), b, SP_REF (c));
+
+      NEXT (2);
+    }
+
   VM_DEFINE_OP (163, unused_163, NULL, NOP)
   VM_DEFINE_OP (164, unused_164, NULL, NOP)
   VM_DEFINE_OP (165, unused_165, NULL, NOP)
