@@ -1,6 +1,6 @@
 ;;; Tree-IL partial evaluator
 
-;; Copyright (C) 2011-2014 Free Software Foundation, Inc.
+;; Copyright (C) 2011-2014, 2017, 2019, 2020 Free Software Foundation, Inc.
 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -144,10 +144,8 @@
         (fold (lambda (name sym res)
                 (vhash-consq sym (make-var name sym 0 #f) res))
               res names gensyms))
-       (($ <letrec> src in-order? names gensyms vals body)
-        (fold (lambda (name sym res)
-                (vhash-consq sym (make-var name sym 0 #f) res))
-              res names gensyms))
+       (($ <letrec>)
+        (error "unexpected letrec"))
        (($ <fix> src names gensyms vals body)
         (fold (lambda (name sym res)
                 (vhash-consq sym (make-var name sym 0 #f) res))
@@ -389,7 +387,7 @@ top-level bindings from ENV and return the resulting expression."
     (let ()
       (define (env-folder x env)
         (match x
-          (($ <toplevel-define> _ name)
+          (($ <toplevel-define> _ _ name)
            (vhash-consq name #t env))
           (($ <seq> _ head tail)
            (env-folder tail (env-folder head env)))
@@ -507,22 +505,14 @@ top-level bindings from ENV and return the resulting expression."
     (define (apply-primitive name args)
       ;; todo: further optimize commutative primitives
       (catch #t
-        (lambda ()
-          (call-with-values
-              (lambda ()
-                (case name
-                  ((eq? eqv?)
-                   ;; Constants will be deduplicated later, but eq?
-                   ;; folding can happen now.  Anticipate the
-                   ;; deduplication by using equal? instead of eq?.
-                   ;; Same for eqv?.
-                   (apply equal? args))
-                  (else
-                   (apply (module-ref the-scm-module name) args))))
-            (lambda results
-              (values #t results))))
-        (lambda _
-          (values #f '()))))
+             (lambda ()
+               (call-with-values
+                   (lambda ()
+                     (apply (module-ref the-scm-module name) args))
+                 (lambda results
+                   (values #t results))))
+             (lambda _
+               (values #f '()))))
     (define (make-values src values)
       (match values
         ((single) single)               ; 1 value
@@ -592,10 +582,6 @@ top-level bindings from ENV and return the resulting expression."
          (let ((body (loop body)))
            (and body
                 (make-let src names gensyms vals body))))
-        (($ <letrec> src in-order? names gensyms vals body)
-         (let ((body (loop body)))
-           (and body
-                (make-letrec src in-order? names gensyms vals body))))
         (($ <fix> src names gensyms vals body)
          (let ((body (loop body)))
            (and body
@@ -716,7 +702,7 @@ top-level bindings from ENV and return the resulting expression."
   (let loop ((exp   exp)
              (env   vlist-null)         ; vhash of gensym -> <operand>
              (counter #f)               ; inlined call stack
-             (ctx 'values))   ; effect, value, values, test, operator, or call
+             (ctx 'values)) ; effect, value, values, test, operator, or call
     (define (lookup var)
       (cond 
        ((vhash-assq var env) => cdr)
@@ -980,29 +966,12 @@ top-level bindings from ENV and return the resulting expression."
                             (lambda (names gensyms vals body)
                               (if (null? names) (error "what!" names))
                               (make-let src names gensyms vals body)))))))
-      (($ <letrec> src in-order? names gensyms vals body)
+      (($ <fix> src names gensyms vals body)
        ;; Note the difference from the `let' case: here we use letrec*
        ;; so that the `visit' procedure for the new operands closes over
        ;; an environment that includes the operands.  Also we don't try
        ;; to elide aliases, because we can't sensibly reduce something
        ;; like (letrec ((a b) (b a)) a).
-       (letrec* ((visit (lambda (exp counter ctx)
-                          (loop exp env* counter ctx)))
-                 (vars (map lookup-var gensyms))
-                 (new (fresh-gensyms vars))
-                 (ops (make-bound-operands vars new vals visit))
-                 (env* (fold extend-env env gensyms ops))
-                 (body* (visit body counter ctx)))
-         (if (and (const? body*) (every constant-expression? vals))
-             ;; We may have folded a loop completely, even though there
-             ;; might be cyclical references between the bound values.
-             ;; Handle this degenerate case specially.
-             body*
-             (prune-bindings ops in-order? body* counter ctx
-                             (lambda (names gensyms vals body)
-                               (make-letrec src in-order?
-                                            names gensyms vals body))))))
-      (($ <fix> src names gensyms vals body)
        (letrec* ((visit (lambda (exp counter ctx)
                           (loop exp env* counter ctx)))
                  (vars (map lookup-var gensyms))
@@ -1043,7 +1012,7 @@ top-level bindings from ENV and return the resulting expression."
                    (else #f))))
                (_ #f))
              (make-let-values lv-src producer (for-tail consumer)))))
-      (($ <toplevel-ref> src (? effect-free-primitive? name))
+      (($ <toplevel-ref> src mod (? effect-free-primitive? name))
        exp)
       (($ <toplevel-ref>)
        ;; todo: open private local bindings.
@@ -1061,10 +1030,10 @@ top-level bindings from ENV and return the resulting expression."
        exp)
       (($ <module-set> src mod name public? exp)
        (make-module-set src mod name public? (for-value exp)))
-      (($ <toplevel-define> src name exp)
-       (make-toplevel-define src name (for-value exp)))
-      (($ <toplevel-set> src name exp)
-       (make-toplevel-set src name (for-value exp)))
+      (($ <toplevel-define> src mod name exp)
+       (make-toplevel-define src mod name (for-value exp)))
+      (($ <toplevel-set> src mod name exp)
+       (make-toplevel-set src mod name (for-value exp)))
       (($ <primitive-ref>)
        (case ctx
          ((effect) (make-void #f))
@@ -1104,12 +1073,6 @@ top-level bindings from ENV and return the resulting expression."
             (make-let src* names vars vals
                       (simplify-conditional
                        (make-conditional src body subsequent alternate))))
-           (($ <conditional> src
-               ($ <letrec> src* in-order? names vars vals body)
-               subsequent alternate)
-            (make-letrec src* in-order? names vars vals
-                         (simplify-conditional
-                          (make-conditional src body subsequent alternate))))
            (($ <conditional> src ($ <fix> src* names vars vals body)
                subsequent alternate)
             (make-fix src* names vars vals
@@ -1193,7 +1156,7 @@ top-level bindings from ENV and return the resulting expression."
               (make-primcall src 'thunk? (list u))
               (make-call src w '())
               (make-primcall
-               src 'scm-error
+               src 'throw
                (list
                 (make-const #f 'wrong-type-arg)
                 (make-const #f "dynamic-wind")
@@ -1219,6 +1182,19 @@ top-level bindings from ENV and return the resulting expression."
                                   (make-call src thunk '())
                                   (make-primcall src 'pop-fluid '()))))))))
 
+      (($ <primcall> src 'with-dynamic-state (state thunk))
+       (for-tail
+        (with-temporaries
+         src (list state thunk) 1 constant-expression?
+         (match-lambda
+          ((state thunk)
+           (make-seq src
+                     (make-primcall src 'push-dynamic-state (list state))
+                     (make-begin0 src
+                                  (make-call src thunk '())
+                                  (make-primcall src 'pop-dynamic-state
+                                                 '()))))))))
+
       (($ <primcall> src 'values exps)
        (cond
         ((null? exps)
@@ -1364,23 +1340,59 @@ top-level bindings from ENV and return the resulting expression."
                (for-tail (make-seq src k (make-const #f #f))))
               (else
                (make-primcall src name (list k (make-const #f elts))))))))
-         (((? equality-primitive?)
-           ($ <lexical-ref> _ _ sym) ($ <lexical-ref> _ _ sym))
-          (for-tail (make-const #f #t)))
 
-         (('= ($ <primcall> src2 'logand (a b)) ($ <const> _ 0))
-          (let ((src (or src src2)))
-            (make-primcall src 'not
-                           (list (make-primcall src 'logtest (list a b))))))
+         (((? equality-primitive?) a (and b ($ <const> _ v)))
+          (cond
+           ((const? a)
+            ;; Constants will be deduplicated later, but eq? folding can
+            ;; happen now.  Anticipate the deduplication by using equal?
+            ;; instead of eq? or eqv?.
+            (for-tail (make-const src (equal? (const-exp a) v))))
+           ((eq? name 'eq?)
+            ;; Already in a reduced state.
+            (make-primcall src 'eq? (list a b)))
+           ((or (memq v '(#f #t () #nil)) (symbol? v) (char? v)
+                (and (exact-integer? v)
+                     (<= most-negative-fixnum v most-positive-fixnum)))
+            ;; Reduce to eq?.  Note that in Guile, characters are
+            ;; comparable with eq?.
+            (make-primcall src 'eq? (list a b)))
+           ((number? v)
+            ;; equal? and eqv? on non-fixnum numbers is the same as
+            ;; eqv?, and can't be reduced beyond that.
+            (make-primcall src 'eqv? (list a b)))
+           ((eq? name 'eqv?)
+            ;; eqv? on anything else is the same as eq?.
+            (make-primcall src 'eq? (list a b)))
+           (else
+            ;; FIXME: inline a specialized implementation of equal? for
+            ;; V here.
+            (make-primcall src name (list a b)))))
+         (((? equality-primitive?) (and a ($ <const>)) b)
+          (for-tail (make-primcall src name (list b a))))
+         (((? equality-primitive?) ($ <lexical-ref> _ _ sym)
+                                   ($ <lexical-ref> _ _ sym))
+          (for-tail (make-const src #t)))
 
          (('logbit? ($ <const> src2
                        (? (lambda (bit)
-                            (and (exact-integer? bit) (not (negative? bit))))
+                            (and (exact-integer? bit)
+                                 (<= 0 bit (logcount most-positive-fixnum))))
                           bit))
                     val)
-          (fold-constants src 'logtest
-                          (list (make-const (or src2 src) (ash 1 bit)) val)
-                          ctx))
+          (for-tail
+           (make-primcall src 'logtest
+                          (list (make-const src2 (ash 1 bit)) val))))
+
+         (('logtest a b)
+          (for-tail
+           (make-primcall
+            src
+            'not
+            (list
+             (make-primcall src 'eq?
+                            (list (make-primcall src 'logand (list a b))
+                                  (make-const src 0)))))))
 
          (((? effect-free-primitive?) . args)
           (fold-constants src name args ctx))
@@ -1467,7 +1479,7 @@ top-level bindings from ENV and return the resulting expression."
                                             opt-vals)))))
 
               (cond
-               ((or (< nargs nreq) (and (not rest) (> nargs (+ nreq nopt))))
+               ((or (< nargs nreq) (and (null? rest) (> nargs (+ nreq nopt))))
                 ;; An error, or effecting arguments.
                 (make-call src (for-call orig-proc) (map for-value orig-args)))
                ((or (and=> (find-counter key counter) counter-recursive?)
@@ -1572,11 +1584,15 @@ top-level bindings from ENV and return the resulting expression."
          (and (not opt) rest (not kw)
               (match body
                 (($ <primcall> _ 'apply
-                    (($ <lambda> _ _ (and lcase ($ <lambda-case>)))
+                    (($ <lambda> _ _ (and lcase ($ <lambda-case> _ req1)))
                      ($ <lexical-ref> _ _ sym)
                      ...))
                  (and (equal? sym gensyms)
                       (not (lambda-case-alternate lcase))
+                      (<= (length req) (length req1))
+                      (every (lambda (s)
+                               (= (lexical-refcount s) 1))
+                             sym)
                       lcase))
                 (_ #f))))
        (let* ((vars (map lookup-var gensyms))
